#include "QTree.h"
#include "UUtil.h"

#define MIN( A, B ) ((A)>(B)?(B):(A))
#define MAX( A, B ) ((A)>(B)?(A):(B))

//
// constructor
//
QTreeNode::QTreeNode( LPDIRECT3DDEVICE9 pd3dDevice, int depth, int left, int right, int bottom, int top, MapVertex *pMapVertices, int rowOffset, std::vector<QTreeNode *>* nodeList)
: m_depth( depth ), m_left( left ), m_right( right ), m_top( top ), m_bottom( bottom ), m_pnodeList( nodeList ), m_patch(0), m_idx( nodeList->size() ), m_aabb(0), m_isLeaf( false )
#ifdef EDITOR_BUILD
, m_toBeUpdated( false )
#endif
{
    // ڽ node list  ߰Ѵ.
    m_pnodeList->push_back( this );

    // depth ʹ ų   ũⰡ ʹ ۴ٸ
    //  ̻  ʴ´.
    if( m_depth > QTree::MAX_DEPTH || right-left < QTree::MIN_PATCH_SIZE || top-bottom < QTree::MIN_PATCH_SIZE )
    {
        m_isLeaf = true;

        // patch Ѵ.
        m_patch = new MapPatch( pd3dDevice, left, right, bottom, top, pMapVertices, rowOffset );

        // AABB Ѵ.
        // vertex  Ѵ.
        createAABB( pMapVertices, rowOffset );
    }
    else
    {
        // child node Ѵ.
        int childDepth = m_depth+1;
        int centerX = ( left + right ) / 2;
        int centerY = ( top + bottom ) / 2;
        // NW
        QTreeNode *nw = new QTreeNode( pd3dDevice, childDepth, left, centerX, centerY, top, pMapVertices, rowOffset, nodeList );
        // NE
        QTreeNode *ne = new QTreeNode( pd3dDevice, childDepth, centerX, right, centerY, top, pMapVertices, rowOffset, nodeList );
        // SE
        QTreeNode *se = new QTreeNode( pd3dDevice, childDepth, centerX, right, bottom, centerY, pMapVertices, rowOffset, nodeList );
        // SW
        QTreeNode *sw = new QTreeNode( pd3dDevice, childDepth, left, centerX, bottom, centerY, pMapVertices, rowOffset, nodeList );

        // child index Ѵ.
        m_child[ NW ] = nw->getIdx();
        m_child[ NE ] = ne->getIdx();
        m_child[ SE ] = se->getIdx();
        m_child[ SW ] = sw->getIdx();        

        // child node ̿ AABB Ѵ.
        AABB3* nwAABB = nw->getAABB();
        AABB3* neAABB = ne->getAABB();
        AABB3* seAABB = se->getAABB();
        AABB3* swAABB = sw->getAABB();

        createAABB( nwAABB, neAABB, seAABB, swAABB );        
    }
}

//
// create a AABB with vertices
//
inline void QTreeNode::createAABB( MapVertex *mapVertices, int rowOffset )
{
    // use a infomations of vertices
    float minY = mapVertices[ m_bottom*rowOffset + m_left ].position.y;
    float maxY = minY;
    float yVal;

    for( int i = m_bottom; i <= m_top; i++ )
    {
        for( int j = m_left; j <= m_right; j++ )
        {
            yVal = mapVertices[ i*rowOffset + j ].position.y;
            if( yVal > maxY )
            {
                maxY = yVal;
            }
            if( yVal < minY )
            {
                minY = yVal;
            }
        }
    }
    if( m_aabb )
    {
        m_aabb->setAABB( (float)m_left, (float)m_right, minY, maxY, (float)m_bottom, (float)m_top );
    }
    else
    {
        m_aabb = new AABB3( (float)m_left, (float)m_right, minY, maxY, (float)m_bottom, (float)m_top );
    }
}

//
// create a AABB with children
//
inline void QTreeNode::createAABB( AABB3 *nwAABB, AABB3 *neAABB, AABB3 *seAABB, AABB3 *swAABB )
{
    float minY, maxY;
    minY = MIN( MIN( neAABB->getMinY(), nwAABB->getMinY() ), MIN( seAABB->getMinY(), swAABB->getMinY() ) );
    maxY = MAX( MAX( neAABB->getMaxY(), nwAABB->getMaxY() ), MAX( seAABB->getMaxY(), swAABB->getMaxY() ) );

    if( m_aabb )
    {
        m_aabb->setAABB( (float)m_left, (float)m_right, minY, maxY, (float)m_bottom, (float)m_top );
    }
    else
    {
        m_aabb = new AABB3( (float)m_left, (float)m_right, minY, maxY, (float)m_bottom, (float)m_top );
    }
}

//
// release
//
void QTreeNode::release()
{
    m_pnodeList = 0;
    Delete( m_patch );
    Delete( m_aabb );
}

//
// frustum culling ϸ鼭 renderingѴ.
//
void QTreeNode::render( UFrustum &frustum )
{
    // intersection test
    CLIPTYPE ret = frustum.check( *m_aabb );
    if( ret == FOUTSIDE )
    {
        return;
    }

    // leaf node̸ ׸.
    if( m_isLeaf )
    {
        m_patch->render( );
    }
    else
    {
        // recursiveϰ child testѴ.
        QTreeNode *child = 0;
        for( int i = 0; i < 4; i ++ )
        {
            child = m_pnodeList->at( m_child[i] );
            child->render( frustum );
        }
    }
}

//
// pickup ray intersectϴ ǥ ãƳ.
//
bool QTreeNode::findPickedPos( const D3DXVECTOR3 *vPickOrg, const D3DXVECTOR3 *vPickDir, D3DXVECTOR3 *vPickOut, float *t, const MapVertex *mapVertices, int rowOffset )
{
    if( m_aabb->intersect( vPickOrg, vPickDir ) )
    {
        if( m_isLeaf )
        {
            return findIntersectedPos( vPickOrg, vPickDir, vPickOut, t, mapVertices, rowOffset );
        }
        else
        {
            // recurse down to children
            bool ret = false;
            QTreeNode *nw = m_pnodeList->at(m_child[NW]);
            QTreeNode *ne = m_pnodeList->at(m_child[NE]);
            QTreeNode *se = m_pnodeList->at(m_child[SE]);
            QTreeNode *sw = m_pnodeList->at(m_child[SW]);
            ret |= nw->findPickedPos( vPickOrg, vPickDir, vPickOut, t, mapVertices, rowOffset );
            ret |= ne->findPickedPos( vPickOrg, vPickDir, vPickOut, t, mapVertices, rowOffset );
            ret |= se->findPickedPos( vPickOrg, vPickDir, vPickOut, t, mapVertices, rowOffset );
            ret |= sw->findPickedPos( vPickOrg, vPickDir, vPickOut, t, mapVertices, rowOffset );

            return ret;
        }
    }

    return false;
}

//
// pickup ray intersectϴ ǥ Ѵ.
//
bool QTreeNode::findIntersectedPos( const D3DXVECTOR3 *vPickOrig, const D3DXVECTOR3 *vPickDir, D3DXVECTOR3 *vPickOut, float *tmpt, const MapVertex *mapVertex, int rowOffset )
{
    const D3DXVECTOR3 *v0, *v1, *v2;
    float t, u, v;
    bool found = false;
    for( int i=m_bottom; i<m_top; i++ )
    {
        for( int j=m_left; j<m_right; j++ )
        {
            // -----
            // |\ 2|
            // | \ |
            // |1 \|
            // -----
            v0 = &(mapVertex[ i*rowOffset + j ].position);
            v1 = &(mapVertex[ (i+1)*rowOffset + j ].position);
            v2 = &(mapVertex[ (i)*rowOffset + j+1 ].position );
            if( D3DXIntersectTri( v0, v1, v2, vPickOrig, vPickDir, &u, &v, &t ) )
            {
                if( *tmpt > t && t >= 0)
                {
                    *tmpt = t;
                    *vPickOut = (*vPickOrig) + (*vPickDir)*t;
                    found = true;
                }
            }

            v0 = &(mapVertex[ i*rowOffset + j+1 ].position);
            v1 = &(mapVertex[ (i+1)*rowOffset + j ].position);
            v2 = &(mapVertex[ (i+1)*rowOffset + j+1 ].position );
            if( D3DXIntersectTri( v0, v1, v2, vPickOrig, vPickDir, &u, &v, &t ) )
            {
                if( *tmpt > t && t >= 0)
                {
                    *tmpt = t;
                    *vPickOut = (*vPickOrig) + (*vPickDir)*t;
                    found = true;
                }
            }
        }
    }

    return found;
}

#ifdef EDITOR_BUILD
//
// 콺 ̿ ִ vertex ã´.
//
void QTreeNode::findPickedVertex( const D3DXVECTOR3 *vPickOrg, const D3DXVECTOR3 *vPickDir, int *idx, float *t, const MapVertex *mapVertices, int rowOffset )
{
    // AABB intersection test
    if( m_aabb->intersect( vPickOrg, vPickDir ) )
    {
        // leaf node̸
        if( m_isLeaf )
        {
            // intersect ϴ vertex ã´.
            findIntersectedVertex( vPickOrg, vPickDir, t, idx, mapVertices, rowOffset );
        }
        else
        {
            // middle node̸
            // child  testѴ.
            QTreeNode *nw = m_pnodeList->at(m_child[NW]);
            QTreeNode *ne = m_pnodeList->at(m_child[NE]);
            QTreeNode *se = m_pnodeList->at(m_child[SE]);
            QTreeNode *sw = m_pnodeList->at(m_child[SW]);
            nw->findPickedVertex( vPickOrg, vPickDir, idx, t, mapVertices, rowOffset );
            ne->findPickedVertex( vPickOrg, vPickDir, idx, t, mapVertices, rowOffset );
            se->findPickedVertex( vPickOrg, vPickDir, idx, t, mapVertices, rowOffset );
            sw->findPickedVertex( vPickOrg, vPickDir, idx, t, mapVertices, rowOffset );
        }
    }
}

//
// find a nearest vertex
//
void QTreeNode::findIntersectedVertex( const D3DXVECTOR3 *vPickOrig, const D3DXVECTOR3 *vPickDir, float *tmpt, int *pickedIdx, const MapVertex *mapVertex, int row )
{
    const D3DXVECTOR3 *v0, *v1, *v2;
    float t, u, v;
    for( int i=m_bottom; i<m_top; i++ )
    {
        for( int j=m_left; j<m_right; j++ )
        {
            v0 = &(mapVertex[ i*row + j ].position);
            v1 = &(mapVertex[ (i+1)*row + j ].position);
            v2 = &(mapVertex[ (i)*row + j+1 ].position );
            if( D3DXIntersectTri( v0, v1, v2, vPickOrig, vPickDir, &u, &v, &t ) )
            {
                if( *tmpt > t && t >= 0)
                {
                    *tmpt = t;
                    D3DXVECTOR3 pickPos = (*vPickOrig) + (*vPickDir)*t;
                    D3DXVECTOR3 e0, e1, e2;
                    D3DXVec3Subtract( &e0, v0, &pickPos );
                    D3DXVec3Subtract( &e1, v1, &pickPos );
                    D3DXVec3Subtract( &e2, v2, &pickPos );
                    float d0 = D3DXVec3Length( &e0 );
                    float d1 = D3DXVec3Length( &e1 );
                    float d2 = D3DXVec3Length( &e2 );
                    if( d0 < d1 )
                    {
                        if( d0 < d2 )
                        {
                            *pickedIdx = i*row+j;
                        }
                        else
                        {
                            *pickedIdx = i*row+j+1;
                        }
                    }
                    else
                    {
                        if( d1 < d2 )
                        {
                            *pickedIdx = (i+1)*row+j;
                        }
                        else
                        {
                            *pickedIdx = i*row+j+1;
                        }
                    }
                }
            }

            v0 = &(mapVertex[ i*row + j+1 ].position);
            v1 = &(mapVertex[ (i+1)*row + j ].position);
            v2 = &(mapVertex[ (i+1)*row + j+1 ].position );
            if( D3DXIntersectTri( v0, v1, v2, vPickOrig, vPickDir, &u, &v, &t ) )
            {
                if( *tmpt > t && t >= 0)
                {
                    *tmpt = t;
                    D3DXVECTOR3 pickPos = (*vPickOrig) + (*vPickDir)*t;
                    D3DXVECTOR3 e0, e1, e2;
                    D3DXVec3Subtract( &e0, v0, &pickPos );
                    D3DXVec3Subtract( &e1, v1, &pickPos );
                    D3DXVec3Subtract( &e2, v2, &pickPos );
                    float d0 = D3DXVec3Length( &e0 );
                    float d1 = D3DXVec3Length( &e1 );
                    float d2 = D3DXVec3Length( &e2 );
                    if( d0 < d1 )
                    {
                        if( d0 < d2 )
                        {
                            *pickedIdx = i*row+j+1;
                        }
                        else
                        {
                            *pickedIdx = (i+1)*row+j+1;
                        }
                    }
                    else
                    {
                        if( d1 < d2 )
                        {
                            *pickedIdx = (i+1)*row+j;
                        }
                        else
                        {
                            *pickedIdx = (i+1)*row+j+1;
                        }
                    }
                }
            }
        }
    }
}

//
// update ɼ ִ AABB üũѴ.
//
bool QTreeNode::updateAABB( int x, int y, float oldYVal, float newYVal )
{
    // update  ǥ  忡 ϴ ˾ƺ.
    if( m_left <= x && x <= m_right && m_bottom <= y && y <= m_top )
    {
        //  
        if( m_isLeaf )
        {
            // update ʿ ȮѴ.
            if( oldYVal == m_aabb->getMinY()
                || oldYVal == m_aabb->getMaxY() 
                || newYVal < m_aabb->getMinY()
                || newYVal > m_aabb->getMaxY() )
            {
                m_toBeUpdated = true;
            }

        }
        // else
        else
        {
            // ڽ  ȮѴ.
            QTreeNode *nw = m_pnodeList->at( m_child[NW] );
            QTreeNode *ne = m_pnodeList->at( m_child[NE] );
            QTreeNode *se = m_pnodeList->at( m_child[SE] );
            QTreeNode *sw = m_pnodeList->at( m_child[SW] );
            m_toBeUpdated |= nw->updateAABB( x, y, oldYVal, newYVal );
            m_toBeUpdated |= ne->updateAABB( x, y, oldYVal, newYVal );
            m_toBeUpdated |= se->updateAABB( x, y, oldYVal, newYVal );
            m_toBeUpdated |= sw->updateAABB( x, y, oldYVal, newYVal );
        }
    }

    return m_toBeUpdated;
}

//
// AABB  ȮѴ.
//
bool QTreeNode::commitAABBAndPatch( MapVertex *mapVertices, int rowOffset )
{
    // update Ȯ ʿϴٸ
    if( m_toBeUpdated )
    {
        // leaf ̸
        if( m_isLeaf )
        {
            // patch  updateѴ.
            m_patch->update( m_left, m_right, m_bottom, m_top, mapVertices, rowOffset );
            // AABB  Ѵ.
            AABB3 oldAABB = *m_aabb;
            createAABB( mapVertices, rowOffset );
            m_toBeUpdated = false;
            //   AABB ٸٸ update Ǿٰ true returnѴ.
            if( oldAABB != *m_aabb )
            {
                return true;
            }
        }
        else
        {
            // middle ̸
            bool needUpdate = false;
            QTreeNode *nw = m_pnodeList->at( m_child[NW] );
            QTreeNode *ne = m_pnodeList->at( m_child[NE] );
            QTreeNode *se = m_pnodeList->at( m_child[SE] );
            QTreeNode *sw = m_pnodeList->at( m_child[SW] );

            // ڽĵ üũѴ.
            needUpdate |= nw->commitAABBAndPatch( mapVertices, rowOffset );
            needUpdate |= ne->commitAABBAndPatch( mapVertices, rowOffset );
            needUpdate |= se->commitAABBAndPatch( mapVertices, rowOffset );
            needUpdate |= sw->commitAABBAndPatch( mapVertices, rowOffset );

            // ڽĵ Ǿٸ
            if( needUpdate )
            {
                // ڽĵ AABB ̿Ͽ  AABB Ѵ.
                AABB3 oldAABB = *m_aabb;
                createAABB( nw->getAABB(), ne->getAABB(), se->getAABB(), sw->getAABB() );
                m_toBeUpdated = false;
                //   AABB ٸ update Ǿٰ true returnѴ.
                if( oldAABB != *m_aabb )
                {
                    return true;
                }
            }
        }
    }
    return false;
}

#endif

//
// quad tree Ѵ.
//
void QTree::create( LPDIRECT3DDEVICE9 pd3dDevice, int row, int col, MapVertex* pMapVertices )
{
    //  ڿ ȯѴ.
    release();

    // root node Ѵ.
    // recursive ϰ  child node constructor Ѵ.
    new QTreeNode( pd3dDevice, 0, 0, row-1, 0, col-1, pMapVertices, row, &m_nodeList );

    m_pMapVertices = pMapVertices;
    m_rowOffset = row;
}

//
// ڿ Ѵ.
//
void QTree::release()
{
    //  node Ѵ.
    for( std::vector<QTreeNode*>::iterator i = m_nodeList.begin(); i != m_nodeList.end(); i++ )
    {
        Delete( *i );
    }
    //  Ѵ.
    m_nodeList.clear();
    
    // UMap  ϹǷ link Ѵ.
    m_pMapVertices = 0;
}

//
// frustum cullingϸ鼭 renderѴ.
//
void QTree::render( UFrustum &frustum )
{
    if( m_nodeList.size() > 0 )
    {
        // root  frustum culling testѴ.
        m_nodeList[0]->render( frustum );
    }
}

//
// 콺 Ŭ  ǥ ˷ش.
//
bool QTree::findPickedPos( const D3DXVECTOR3 *vPickDir, const D3DXVECTOR3 *vPickOrg, D3DXVECTOR3 *vOut )
{
    // Map   false return Ѵ.
    if( m_nodeList.size() == 0 )
    {
        return false;
    }

    // root node ã Ѵ.
    float t = 2000;
    return m_nodeList[0]->findPickedPos( vPickOrg, vPickDir, vOut, &t, m_pMapVertices, m_rowOffset );
}

#ifdef EDITOR_BUILD
//
// mouse  ִ vertex ãƳ.
// ã ϸ -1 returnѴ.
//
int QTree::findPickedVertex( const D3DXVECTOR3 *vPickDir, const D3DXVECTOR3 *vPickOrg )
{
    //  ִٸ -1 returnѴ.
    if( m_nodeList.size() == 0 )
    {
        return -1;
    }

    int idx = -1;
    float t = 2000;
    // root node ã Ѵ.
    m_nodeList[0]->findPickedVertex( vPickOrg, vPickDir, &idx, &t, m_pMapVertices, m_rowOffset );

    return idx;
}

//
// update ɼ ִ AABB üũѴ.
//
void QTree::updateAABB( int x, int y, float oldYVal, float newYVal )
{
    if( m_nodeList.size() > 0 )
    {
        // root node  üũϱ Ѵ.
        m_nodeList[0]->updateAABB( x, y, oldYVal, newYVal );
    }
}

//
// AABB Ѵ.
//
void QTree::commitAABBAndPatch()
{
    if( m_nodeList.size() > 0 )
    {
        // root node  üũϱ Ѵ.
        m_nodeList[0]->commitAABBAndPatch( m_pMapVertices, m_rowOffset );
    }
}
#endif