#include "UCamera.h"
#include "udebug.h"

// define this when using y+ vector for a up vector of a camera
//#define USE_YUP_VECTOR

const float UCamera::MIN_DISTANCE = 3.0f;
const float UCamera::MAX_DISTANCE = 10.0f;
const float UCamera::MAX_YANGLE = D3DX_PI*2/5;
const float UCamera::MIN_YANGLE = 0;

D3DXVECTOR3 UCamera::f0(-1.0f,-1.0f, 1.0f);
D3DXVECTOR3 UCamera::f1(-1.0f, 1.0f, 1.0f);
D3DXVECTOR3 UCamera::f2( 1.0f, 1.0f, 1.0f);
D3DXVECTOR3 UCamera::f3( 1.0f,-1.0f, 1.0f);

D3DXVECTOR3 UCamera::vY( 0, 1.0f, 0 );

//
// initialize camera
// 
//   ϸ y+ 45  z+    ó view transformation
// matrix Ѵ. ׸ fovy : D3DX_PI/4 , 1.0f, near : 1.0f, far :1000.0f
//  projection matrix Ѵ.
//
// 1. x-y 鿡
//    eye = ( r * cos A, r * sin A, 0 )
//    up =  ( cos 90+A, sin 90+A, 0 )
// 2. y   rotate
//    eye = ( r * cos A * cos B, r * sin A, -r * cosA * sin B )
//    up  = ( cos 90+A *cos B, sin 90+A,  cos 90+A * sin B )
// 3. lookat ŭ trans formation
//    eye + lookat
//    lookat
//    up
//
void UCamera::init( D3DXVECTOR3 lookat )
{
    m_distance = 10.0f;
    m_yAngle = D3DX_PI/4;
    m_zAngle = D3DX_PI*3/2;

    m_cosA = cosf( m_yAngle );
    m_sinA = sinf( m_yAngle );
    m_cos90A = cosf( m_yAngle + D3DX_PI/2 );
    m_sin90A = sinf( m_yAngle + D3DX_PI/2 );
    m_cosB = cosf( m_zAngle );
    m_sinB = sinf( m_zAngle );

    // calculate view transformation matrix
    m_lookat = lookat;
    m_eyeNormal.x = m_cosA * m_cosB;
    m_eyeNormal.y = m_sinA;
    m_eyeNormal.z = m_cosA * m_sinB;
    m_eye = m_eyeNormal*m_distance + m_lookat;
#ifdef USE_YUP_VECTOR
    m_up = vY;
#else
    m_up.x = m_cos90A * m_cosB;
    m_up.y = m_sin90A;
    m_up.z = m_cos90A * m_sinB;
#endif

    D3DXMatrixLookAtLH( &m_matView, &m_eye, &m_lookat, &m_up );

    // calcuate the direction vector
    calcDirVec();

    // calculate projection matrix
    m_fovy = D3DX_PI/4;
    m_aspect = 1.0f;
    m_near = 1.0f;
    m_far = 100.0f;

    D3DXMatrixPerspectiveFovLH( &m_matProj, m_fovy, m_aspect, m_near, m_far );

    // calculate the view-projection matrix
    D3DXMATRIX comboMatrix;
    D3DXMatrixMultiply( &comboMatrix, &m_matView, &m_matProj );

    // generate frustum
    m_frustum.init( comboMatrix );

    calcFrustumOnXZ();
}

//
// calculate direction vector
//
inline void UCamera::calcDirVec()
{
    // calculate view direction and left direction
    // find direction
    m_vDir = m_lookat - m_eye;
    // project to x-z plane
    m_vDir.y = 0;
    // normalize
    D3DXVec3Normalize( &m_vDir, &m_vDir );

    // find left direction
    D3DXVec3Cross( &m_vLDir, &m_vDir, &vY);
    // normalize
    D3DXVec3Normalize( &m_vLDir, &m_vLDir );
}

//
// zoom in/out a camera
//
void UCamera::zoom( float fdis )
{
    // calculate new distance
    m_distance += fdis;
    if( m_distance < MIN_DISTANCE )
    {
        m_distance = MIN_DISTANCE;
    }
    if( m_distance > MAX_DISTANCE )
    {
        m_distance = MAX_DISTANCE;
    }

    // recalculate eye position
    m_eye = m_eyeNormal*m_distance + m_lookat;
    // create view matrix
    D3DXMatrixLookAtLH( &m_matView, &m_eye, &m_lookat, &m_up );

    calcDirVec();

    // calculate the view-projection matrix
    D3DXMATRIX comboMatrix;
    D3DXMatrixMultiply( &comboMatrix, &m_matView, &m_matProj );

    // generate frustum
    m_frustum.init( comboMatrix );

    calcFrustumOnXZ();
}

//
// move
//
void UCamera::move( D3DXVECTOR3 newPosition )
{
    // caculate the difference between old position and new position
    D3DXVECTOR3 diff = m_lookat - newPosition;
    // set new position
    m_lookat = newPosition;
    // move eye position with the difference of lookat position
    m_eye -= diff;
    
    // calculate view matrix
    D3DXMatrixLookAtLH( &m_matView, &m_eye, &m_lookat, &m_up );

    // calculate the view-projection matrix
    D3DXMATRIX comboMatrix;
    D3DXMatrixMultiply( &comboMatrix, &m_matView, &m_matProj );

    // generate frustum
    m_frustum.init( comboMatrix );

    calcFrustumOnXZ();
}

//
// move a camera front
//
void UCamera::moveFront( const float delta )
{
    D3DXVECTOR3 vDiff = m_vDir * delta;

    m_lookat += vDiff;
    m_eye += vDiff;
 
    // calculate view matrix
    D3DXMatrixLookAtLH( &m_matView, &m_eye, &m_lookat, &m_up );

    // calculate the view-projection matrix
    D3DXMATRIX comboMatrix;
    D3DXMatrixMultiply( &comboMatrix, &m_matView, &m_matProj );

    // generate frustum
    m_frustum.init( comboMatrix );

    calcFrustumOnXZ();
}

//
// move a camera back
//
void UCamera::moveBack( const float delta )
{
    D3DXVECTOR3 vDiff = m_vDir * delta;

    m_lookat -= vDiff;
    m_eye -= vDiff;
 
    // calculate view matrix
    D3DXMatrixLookAtLH( &m_matView, &m_eye, &m_lookat, &m_up );

    // calculate the view-projection matrix
    D3DXMATRIX comboMatrix;
    D3DXMatrixMultiply( &comboMatrix, &m_matView, &m_matProj );

    // generate frustum
    m_frustum.init( comboMatrix );

    calcFrustumOnXZ();
}

//
// move a camera left
//
void UCamera::moveLeft( const float delta )
{
    D3DXVECTOR3 vDiff = m_vLDir * delta;

    m_lookat += vDiff;
    m_eye += vDiff;
 
    // calculate view matrix
    D3DXMatrixLookAtLH( &m_matView, &m_eye, &m_lookat, &m_up );

    // calculate the view-projection matrix
    D3DXMATRIX comboMatrix;
    D3DXMatrixMultiply( &comboMatrix, &m_matView, &m_matProj );

    // generate frustum
    m_frustum.init( comboMatrix );

    calcFrustumOnXZ();
}

//
// move a camera right
//
void UCamera::moveRight( const float delta )
{
    D3DXVECTOR3 vDiff = m_vLDir * delta;

    m_lookat -= vDiff;
    m_eye -= vDiff;
 
    // calculate view matrix
    D3DXMatrixLookAtLH( &m_matView, &m_eye, &m_lookat, &m_up );

    // calculate the view-projection matrix
    D3DXMATRIX comboMatrix;
    D3DXMatrixMultiply( &comboMatrix, &m_matView, &m_matProj );

    // generate frustum
    m_frustum.init( comboMatrix );

    calcFrustumOnXZ();
}

//
// rotate horizontally
//
void UCamera::rotateHorizontally( float angle )
{
    // calculate new angle
    m_zAngle += angle;
    if( m_zAngle > 2*D3DX_PI )
    {
        m_zAngle -= 2*D3DX_PI;
    }

    m_cosB = cosf( m_zAngle );
    m_sinB = sinf( m_zAngle );

    // new eye position and up vector
    m_eyeNormal.x = m_cosA * m_cosB;
    m_eyeNormal.z = m_cosA * m_sinB;
    m_eye = m_eyeNormal*m_distance + m_lookat;
#ifdef USE_YUP_VECTOR
    m_up = vY;
#else
    m_up.x = m_cos90A * m_cosB;
    m_up.z = m_cos90A * m_sinB;
#endif

    // calculate view matrix
    D3DXMatrixLookAtLH( &m_matView, &m_eye, &m_lookat, &m_up );

    calcDirVec();

    // calculate the view-projection matrix
    D3DXMATRIX comboMatrix;
    D3DXMatrixMultiply( &comboMatrix, &m_matView, &m_matProj );

    // generate frustum
    m_frustum.init( comboMatrix );

    calcFrustumOnXZ();
}

//
// rotate vertically
//
void UCamera::rotateVertically( float angle )
{
    m_yAngle += angle;
    if( m_yAngle > MAX_YANGLE )
    {
        m_yAngle = MAX_YANGLE;
    }
    if ( m_yAngle < MIN_YANGLE )
    {
        m_yAngle = MIN_YANGLE;
    }

    m_cosA = cosf( m_yAngle );
    m_sinA = sinf( m_yAngle );
    m_cos90A = cosf( m_yAngle + D3DX_PI/2 );
    m_sin90A = sinf( m_yAngle + D3DX_PI/2 );

    // calculate view transformation matrix
    m_eyeNormal.x = m_cosA * m_cosB;
    m_eyeNormal.y = m_sinA;
    m_eyeNormal.z = m_cosA * m_sinB;
    m_eye = m_eyeNormal*m_distance + m_lookat;
#ifdef USE_YUP_VECTOR
    m_up = vY;
#else
    m_up.x = m_cos90A * m_cosB;
    m_up.y = m_sin90A;
    m_up.z = m_cos90A * m_sinB;
#endif

    D3DXMatrixLookAtLH( &m_matView, &m_eye, &m_lookat, &m_up );

    calcDirVec();

    // calculate the view-projection matrix
    D3DXMATRIX comboMatrix;
    D3DXMatrixMultiply( &comboMatrix, &m_matView, &m_matProj );

    // generate frustum
    m_frustum.init( comboMatrix );

    calcFrustumOnXZ();
}

//
// rotate horizontally and vertically
//
void UCamera::rotate( float zAngle, float yAngle )
{
    // calculate new z angle
    m_zAngle += zAngle;
    if( m_zAngle > 2*D3DX_PI )
    {
        m_zAngle -= 2*D3DX_PI;
    }

    // calculate new y angle
    m_yAngle += yAngle;
    if( m_yAngle > MAX_YANGLE )
    {
        m_yAngle = MAX_YANGLE;
    }
    if ( m_yAngle < MIN_YANGLE )
    {
        m_yAngle = MIN_YANGLE;
    }

    m_cosA = cosf( m_yAngle );
    m_sinA = sinf( m_yAngle );
    m_cos90A = cosf( m_yAngle + D3DX_PI/2 );
    m_sin90A = sinf( m_yAngle + D3DX_PI/2 );
    m_cosB = cosf( m_zAngle );
    m_sinB = sinf( m_zAngle );

    // calculate view transformation matrix
    m_eyeNormal.x = m_cosA * m_cosB;
    m_eyeNormal.y = m_sinA;
    m_eyeNormal.z = m_cosA * m_sinB;
    m_eye = m_eyeNormal*m_distance + m_lookat;
#ifdef USE_YUP_VECTOR
    m_up = vY;
#else
    m_up.x = m_cos90A * m_cosB;
    m_up.y = m_sin90A;
    m_up.z = m_cos90A * m_sinB;
#endif

    D3DXMatrixLookAtLH( &m_matView, &m_eye, &m_lookat, &m_up );

    calcDirVec();

    // calculate the view-projection matrix
    D3DXMATRIX comboMatrix;
    D3DXMatrixMultiply( &comboMatrix, &m_matView, &m_matProj );

    // generate frustum
    m_frustum.init( comboMatrix );

    calcFrustumOnXZ( );
}

//
// set view transformation matrix
//
void UCamera::setCamera( LPDIRECT3DDEVICE9 pd3dDevice )
{
    pd3dDevice->SetTransform( D3DTS_VIEW, &m_matView );
    pd3dDevice->SetTransform( D3DTS_PROJECTION, &m_matProj );
}

//
// calculate edge vertices of a frustum on X-Z plane
//
void UCamera::calcFrustumOnXZ()
{
    // unproject far plane
    D3DXVECTOR3 vUF[4];

    D3DXVec3Unproject( &(vUF[0]), &f0, NULL, &m_matProj, &m_matView, NULL );
    D3DXVec3Unproject( &(vUF[1]), &f1, NULL, &m_matProj, &m_matView, NULL );
    D3DXVec3Unproject( &(vUF[2]), &f2, NULL, &m_matProj, &m_matView, NULL );
    D3DXVec3Unproject( &(vUF[3]), &f3, NULL, &m_matProj, &m_matView, NULL );
    
    // find intersected point on X-Z plane
    D3DXPLANE xzPlane = D3DXPLANE( 0, 1, 0, 0 );
    D3DXPlaneIntersectLine( &m_edgeVertices[0], &xzPlane, &m_eye, &vUF[0] );
    D3DXPlaneIntersectLine( &m_edgeVertices[1], &xzPlane, &m_eye, &vUF[1] );
    D3DXPlaneIntersectLine( &m_edgeVertices[2], &xzPlane, &m_eye, &vUF[2] );
    D3DXPlaneIntersectLine( &m_edgeVertices[3], &xzPlane, &m_eye, &vUF[3] );
}

//
// calculate a polygon that intersects with a frustum on X-Z plane
//
void UCamera::getFrustumOnXZ( D3DXVECTOR3 *edgeVertices )
{
    memcpy( edgeVertices, m_edgeVertices, 4*sizeof(D3DXVECTOR3) );
}